﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using System.Threading;
using Nansen.StatusCodes.Persistence;
using EPiServer.Core;
using EPiServer;
using EPiServer.DataAbstraction;
using System.Net;
using System.Text.RegularExpressions;

namespace Nansen.StatusCodes.Modules
{
    /// <summary>
    /// Module for providing correct status codes to search engines and visitors alike.
    /// Assumes Friendly URL Provider
    /// </summary>
    public class StatusCodeModule : IHttpModule
    {
        // This can be set to your own implementation of IUrlMapper
        private Type _urlMapperType = typeof(SimpleSqlUrlMapper);
        private static readonly object _lockObject = new object();
        private static bool _initialized = false;

        private IUrlMapper Mapper
        {
            get
            {
                return Activator.CreateInstance(_urlMapperType) as IUrlMapper;
            }
        }

        public void Dispose()
        {
            EPiServer.DataFactory.Instance.DeletedPage -= Instance_DeletedPage;
            EPiServer.DataFactory.Instance.MovedPage -= Instance_MovedPage;
            EPiServer.DataFactory.Instance.PublishedPage -= Instance_MovedPage;

            _initialized = false;
        }

        public void Init(HttpApplication context)
        {
            SetupEvents();

            context.EndRequest += new EventHandler(context_EndRequest);
        }

        private void context_EndRequest(object sender, EventArgs e)
        {
            HttpApplication app = sender as HttpApplication;
            if (app != null && app.Context.Response.StatusCode == (int)HttpStatusCode.NotFound)
            {
                HttpStatusCode? statusCode = this.Mapper.GetStatusCode(app.Context.Request.Url.AbsolutePath);

                if (statusCode != null)
                {
                    app.Context.Server.ClearError();
                    app.Context.Response.Clear();
                    app.Context.Response.StatusCode = (int)statusCode.Value;

                    if (statusCode == HttpStatusCode.MovedPermanently)
                    {
                        app.Context.Response.AddHeader("Location", this.Mapper.GetRedirectUrl(app.Context.Request.Url.AbsolutePath));
                    }
                }
            }
        }

        private void SetupEvents()
        {
            if (!_initialized)
            {
                try
                {
                    if (Monitor.TryEnter(_lockObject))
                    {
                        this.Mapper.GenerateSchema();

                        EPiServer.DataFactory.Instance.DeletedPage += new EPiServer.PageEventHandler(Instance_DeletedPage);
                        EPiServer.DataFactory.Instance.MovedPage += new EPiServer.PageEventHandler(Instance_MovedPage);
                        EPiServer.DataFactory.Instance.PublishedPage += new EPiServer.PageEventHandler(Instance_PublishedPage);

                        _initialized = true;
                    }
                }
                finally
                {
                    Monitor.Exit(_lockObject);
                }
            }
        }

        private void Instance_PublishedPage(object sender, EPiServer.PageEventArgs e)
        {
            string url = GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL);

            PageVersionCollection pvc = PageVersion.List(e.PageLink);

            // Add a mapping if you change the url segment
            if (pvc.Count > 1)
            {
                PageData previousPage = EPiServer.DataFactory.Instance.GetPage(pvc[pvc.Count - 2].ID);
                if(previousPage.URLSegment != EPiServer.DataFactory.Instance.GetPage(e.PageLink).URLSegment)
                {
                    string oldUrl = GetFriendlyURL(previousPage.PageLink, EPiServer.DataFactory.Instance.GetPage(previousPage.PageLink).LinkURL);
                    string newUrl = GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL);

                    this.Mapper.Add301MovedPermanentlyStatus(oldUrl, newUrl);
                }
            }

            if (this.Mapper.Exists(GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL)))
            {
                this.Mapper.RemoveMappings(GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL));
            }
        }

        private void Instance_MovedPage(object sender, EPiServer.PageEventArgs e)
        {
            // HACK: Just remove all the mappings when moving from the waste basket.
            if (EPiServer.DataFactory.Instance.GetPage(e.PageLink).ParentLink == PageReference.WasteBasket)
            {
                this.Mapper.RemoveMappings(GetFriendlyURL(e.TargetLink, EPiServer.DataFactory.Instance.GetPage(e.TargetLink).LinkURL) + EPiServer.DataFactory.Instance.GetPage(e.PageLink).URLSegment + "/");
            }
            else if (e.TargetLink == PageReference.WasteBasket)
            {
                this.Mapper.Add410GoneStatus(GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL));
            }
            else
            {
                string oldUrl = GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL);
                string newUrl = GetFriendlyURL(e.TargetLink, EPiServer.DataFactory.Instance.GetPage(e.TargetLink).LinkURL) + EPiServer.DataFactory.Instance.GetPage(e.PageLink).URLSegment + "/";

                this.Mapper.Add301MovedPermanentlyStatus(oldUrl, newUrl);

                // TODO: When moving a page to a container where another page has the same URLSegment, 
                // EPiServer adds stuff to the urlsegment after the pageevent chain, making the mappings incorrect.
                // In the meantime, just remove all mapping in case that should happen (shouldn't be too often...)
                foreach (PageData pd in EPiServer.DataFactory.Instance.GetChildren(e.TargetLink))
                {
                    if (pd.URLSegment == EPiServer.DataFactory.Instance.GetPage(e.PageLink).URLSegment)
                    {
                        this.Mapper.RemoveMappings(GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL));
                    }
                }
            }
        }

        private void Instance_DeletedPage(object sender, EPiServer.PageEventArgs e)
        {
            this.Mapper.Add410GoneStatus(GetFriendlyURL(e.PageLink, EPiServer.DataFactory.Instance.GetPage(e.PageLink).LinkURL));
        }

        // Shamelessly stolen from tednyberg.com
        private static string GetFriendlyURL(PageReference pageLink, string url)
        {
            UrlBuilder urlBuilder = new UrlBuilder(url);
            EPiServer.Global.UrlRewriteProvider.ConvertToExternal(urlBuilder, pageLink, System.Text.UTF8Encoding.UTF8);
            return urlBuilder.Uri.ToString();
        }
    }
}